/***************************************************************************
                          frontend.h  -  description
                             -------------------
    begin                : Mon Jun 2 2003
    copyright            : (C) 2003 by Elad Lahav
    email                : elad_lahav@users.sf.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FRONTEND_H
#define FRONTEND_H

#include <qobject.h>
#include <kprocess.h>


/**
 * Represents a single token in the parsed output stream.
 * @author Elad Lahav
 */

class FrontendToken
{
public:
	/**
	 * Class constructor.
	 */
	FrontendToken() : m_pNext(NULL) {}

	const QString& getData() const { return m_sData; }
	FrontendToken* getNext() const { return m_pNext; }

protected:
	QString m_sData;
	FrontendToken* m_pNext;

	friend class Frontend;
};

/**
 * Abstract base class for classes that display data obtained from back-end
 * programmes.
 * @author Elad Lahav
 */

class FrontendDataTarget
{
public:
	/**
	 * Class constructor.
	 */
	FrontendDataTarget() {}

	/**
	 * Called by a Frontend object when new data can be read by the target
	 * object.
	 * @param	pToken	Points to the first token in the list of new data
	 */
	virtual void dataReady(FrontendToken* pToken) = 0;

	/**
	 * Called by the Frontend object to notify when an operation has ended, 
	 * and reports the number of records resulted from the operation.
	 * @param	nRecords	number of records returned by the query
	 */
	virtual void finished(int /* nRecords */) {}
};
 
/**
 * Abstract base class that provides a front-end to console-based programmes.
 * Provides a parsing infrastructure which is based on a list of entries, all with
 * the same structure. Each entry is composed of a number of delimited fields.
 * @author Elad Lahav
 */

class Frontend : public QObject
{
	Q_OBJECT
	
public: 
	Frontend(uint);
	~Frontend();

	virtual bool run(const QStringList&,
		const QString& sWorkDir = QString::null,
		FrontendDataTarget* pTarget = NULL);
	virtual void exit();

	/**
	 * Determines whether the given path leads to the executable for which
	 * this class serves as a front-end.
	 * @param	sPath	The path to the executable
	 * @return	true if the path leads to the expected file, false otherwise
	 */
	virtual bool verify(const QString& sPath) = 0;

	inline bool isRunning() { return m_pProc != NULL; }
	
signals:
	void running(bool);

protected:
	enum ParserDelim { Newline = 0x01, Space = 0x02, Tab = 0x04,
		WSpace = Space | Tab, All = WSpace | Newline };

	/** Defines the set of return values for parseStdout(). Determines what
		needs to be done with a new token passed to this method. */
	enum ParseResult { DiscardToken	/** Delete this token */,
		AcceptToken					/** Add this token to the list */,
		EntryReady					/** This token completes a record */
	};

	/** The embedded back-end process */
	KProcess* m_pProc;

	/** The object to receive parsed data from the process output stream */
	FrontendDataTarget* m_pTarget;

	/** The head of the list of parsed output tokens */
	FrontendToken* m_pHeadToken;

	/** The tail of the list of parsed output tokens */
	FrontendToken* m_pTailToken;

	/** An iterator on the list of parsed output tokens */
	FrontendToken* m_pCurToken;

	/** The current delimiters used for parsing the output */
	ParserDelim m_delim;

	void addToken(FrontendToken*);
	void removeToken();
	void removeEntry();
	bool tokenize(char**, int*, QString&, ParserDelim&);

	/**
	 * Handles a text token received on the Standard Output stream of the
	 * controlled process.
	 * This is called by slotReadStdout whenever a new token is recognised.
	 * Inheriting classes should implement this method to parse the resutling
	 * stream of tokens.
	 * @param	sToken	A part of the text received on the Standard Output,
	 *					disected according to current delimiter settings
	 * @param	delim	The delimiter that ended this token
	 * @result	A ParseResult value, indicating what should be done with the
	 *			new token
	 */
	virtual ParseResult parseStdout(QString& sToken, ParserDelim delim) = 0;

	virtual void parseStderr(const QString&);
	virtual void finalize() {}

protected slots:
	virtual void slotProcessExit(KProcess*);
	
private:
	/** Counts the number of running processes (since it may take some time
		for a process to die, while another is executed). This variable is
		used for emitting the running() signal. */
	uint m_nProcRef;

	/** Determines whether the parser is in the middle of a token, or between
		two tokens */
	bool m_bInToken;

	/** The number of fields in each parsed entry. Should be defined for 
		every sub-class. */
	uint m_nEntrySize;
		
private slots:
	void slotReadStdout(KProcess*, char*, int);
	void slotReadStderr(KProcess*, char*, int);
};

#endif
