/***************************************************************************
                          querypage.cpp  -  description
                             -------------------
    begin                : Sat May 24 2003
    copyright            : (C) 2003 by Elad Lahav
    email                : elad_lahav@users.sf.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qfile.h>
#include <klocale.h>
#include "querypage.h"
#include "kscopeconfig.h"
#include "cscopefrontend.h"

#define FILE_VERSION	"VERSION=2"

const char* QUERY_TYPES[][2] = {
	{ "References to ", "REF " },
	{ "Definition of ", "DEF " },
	{ "Functions called by ", "<-- " },
	{ "Functions calling ", "-->" },
	{ "Search for ", "TXT " },
	{ "", "" },
	{ "EGrep Search for ", "GRP " },
	{ "Files named ", "FIL " },
	{ "Files #including ", "INC " }
};

/**
 * Class constructor.
 * @param	pParent	The parent widget
 * @param	szName	The widget's name
 */
QueryPage::QueryPage(QWidget* pParent, const char * szName) :
	QHBox(pParent, szName),
	m_bLocked(false),
	m_nType(CscopeFrontend::None)
{
	// Create the query results list view
	m_pList = new QListView(this);
	m_pList->setAllColumnsShowFocus(true);

	// Initialise the list's columns
	m_pList->addColumn(i18n("File"));
	m_pList->addColumn(i18n("Function"));
	m_pList->addColumn(i18n("Line"));
	m_pList->addColumn(i18n("Text"));
	m_pList->setColumnAlignment(2, Qt::AlignRight);

	// Set colours and font
	applyPrefs();

	// A record is selected if it is either double-clicked, or the "Enter"
	// key is pressed while the record has the focus
	connect(m_pList, SIGNAL(doubleClicked(QListViewItem*)), this, 
		SIGNAL(recordSelected(QListViewItem*)));
	connect(m_pList, SIGNAL(returnPressed(QListViewItem*)), this, 
		SIGNAL(recordSelected(QListViewItem*)));
}

/**
 * Class destructor.
 */
QueryPage::~QueryPage()
{
}

/**
 * Adds a query entry to the list.
 * Called by a CscopeFrontend object, when a new entry was received in its
 * whole from the Cscope back-end process.
 * @param	pToken	The first token in the entry
 */
void QueryPage::dataReady(FrontendToken* pToken)
{
	QString sFile, sFunc, sLine, sText;
	
	// Get the file name
	sFile = pToken->getData();
	pToken = pToken->getNext();

	// Get the function name
	sFunc = pToken->getData();
	pToken = pToken->getNext();

	// Get the line number
	sLine = pToken->getData();
	pToken = pToken->getNext();

	// Get the line's text
	sText = pToken->getData();
	pToken = pToken->getNext();

	// Add a list item
	(void)new QListViewItem(m_pList, sFile, sFunc, sLine, sText);
}

/**
 * Handles a finished query event, reported by the Cscope frontend object.
 * If no resutls are available, a proper message is displayed. If only one 
 * record was generated by Cscope, it is automatically selected for viewing.
 */
void QueryPage::finished(int nRecords)
{
	if (nRecords < 1)
		(void)new QListViewItem(m_pList, "No resutls");
	
	emit queryFinished();
	
	if (nRecords == 1)
		emit recordSelected(m_pList->firstChild());
}

/**
 * Sets the list's colours and font, according the user's preferences.
 */
void QueryPage::applyPrefs()
{
	// Apply colour settings
	m_pList->setPaletteBackgroundColor(Config().getColor(
		KScopeConfig::QueryList, KScopeConfig::Background));
	m_pList->setPaletteForegroundColor(Config().getColor(
		KScopeConfig::QueryList, KScopeConfig::Foreground));
	m_pList->setFont(Config().getFont(KScopeConfig::QueryList));
}

/**
 * Restores a locked query from the given query file.
 * NOTE: The query file is deleted when loading is complete.
 * @param	sProjPath	The full path of the project directory
 * @param	sFileName	The name of the query file to load
 * @return	true if successful, false otherwise
 */
bool QueryPage::load(const QString& sProjPath, const QString& sFileName)
{
	QListViewItem* pItem;
	QString sTemp, sFile, sFunc, sLine, sText;
	int nState;
	
	// Try to open the query file for reading
	QFile file(sProjPath + "/" + sFileName);
	if (!file.open(IO_ReadOnly))
		return false;
	
	{
		// Use a new scope for the QTextStream object, to ensure its 
		// destruction before the file is deleted
		QTextStream str(&file);
		
		// Make sure the file's version is correct
		sTemp = str.readLine();
		if (sTemp != FILE_VERSION) {
			file.remove();
			return false;
		}
		
		// Read the query name
		m_sName = str.readLine();
		if (m_sName == QString::null || m_sName.isEmpty())
			return false;
		
		// Read the query's type
		sTemp = str.readLine();
		if (sTemp == QString::null || sTemp.isEmpty())
			return false;
		
		// Convert the type string to an integer
		m_nType = sTemp.toUInt();
		if (m_nType >= CscopeFrontend::None) {
			m_nType = CscopeFrontend::None;
			return false;
		}		
		
		// Read the query's text
		m_sText = str.readLine();
		if (m_sText == QString::null || m_sText.isEmpty())
			return false;
		
		// Read query records
		sTemp = str.readLine();
		nState = 0;
		while (sTemp != QString::null) {
			switch (nState) {
			// File path
			case 0:
				sFile = sTemp;
				break;
				
			// Function name
			case 1:
				sFunc = sTemp;
				break;
				
			// Line number
			case 2:
				sLine = sTemp;
				break;
				
			// Text string
			case 3:
				sText = sTemp;
				pItem = new QListViewItem(m_pList, sFile, sFunc, sLine, 
					sText);
				break;
			}
			
			nState = (nState + 1) % 4;
			sTemp = str.readLine();
		}
	}
	
	// Delete the query file
	file.remove();
	
	return true;
}

/**
 * Writes the contents of the page to a query file.
 * This method is called for locked queries before the owner project is
 * closed.
 * @param	sProjPath	The full path of the project directory
 * @return	The name of the query file generated for this page (or an empty
 *			string if no file was written)
 */
QString QueryPage::store(const QString& sProjPath)
{
	QString sFileName, sFileNameBase;
	int i = 0;
	QListViewItemIterator itr(m_pList);
	
	// Do nothing if not initialised
	if (m_sName.isEmpty())
		return QString();
	
	// Create a unique file name
	sFileNameBase = m_sName;
	sFileNameBase.replace(' ', '_');
	do {
		sFileName = sFileNameBase + QString::number(++i);
	} while (QFile(sProjPath + "/" + sFileName).exists());
	
	// Open the query file for writing
	QFile file(sProjPath + "/" + sFileName);
	if (!file.open(IO_WriteOnly))
		return QString();
	
	QTextStream str(&file);
	
	// Write the version string
	str << FILE_VERSION << "\n";
	
	// Write the name, type and text of the query
	str << m_sName << "\n" << m_nType << "\n" << m_sText << "\n";
	
	// Write all records
	for(; itr.current(); ++itr) {
		str << itr.current()->text(0) << "\n" 
			<< itr.current()->text(1) << "\n"
			<< itr.current()->text(2) << "\n"
			<< itr.current()->text(3) << "\n";
	}
	
	return sFileName;
}

/**
 * Resets the query page by deleting all records.
 */
void QueryPage::clear()
{
	m_pList->clear();
	m_nType = CscopeFrontend::None;
	m_sText = QString();
	m_sName = QString();
}

/**
 * Associates the page with a query's type and text.
 * @param	nType	The type of the query
 * @param	sText	The text of the query
 */
void QueryPage::setQueryData(uint nType, const QString& sText)
{
	m_nType = nType;
	m_sText = sText;
	m_sName = getQueryCaption();
}

/** 
 * Constructs a caption for this page, based on the query's type and text.
 * @param	bBrief	true to use a shortened version of the caption, false
 *					(default) for the full version
 * @return	The caption for this page
 */
QString QueryPage::getQueryCaption(bool bBrief) const
{
	return QString(QUERY_TYPES[m_nType][bBrief ? 1 : 0] + m_sText);
}
